package com.bergerkiller.bukkit.nolagg;

import java.io.DataInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import org.bukkit.craftbukkit.CraftWorld;

import com.bergerkiller.bukkit.nolagg.bufferedtag.NBTBaseBuffered;
import com.bergerkiller.bukkit.nolagg.datapool.Pool;
import com.bergerkiller.bukkit.nolagg.datapool.PoolBuffer;
import com.bergerkiller.bukkit.nolagg.datapool.PoolData;

import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.Entity;
import net.minecraft.server.EntityTypes;
import net.minecraft.server.NBTTagCompound;
import net.minecraft.server.NBTTagList;
import net.minecraft.server.NibbleArray;
import net.minecraft.server.TileEntity;
import net.minecraft.server.World;
import net.minecraft.server.WorldServer;

public class BufferedChunk extends Chunk implements PoolData<ChunkBuffer> {
	private static Map<ChunkCoord, BufferedChunk> unloadedChunks = new HashMap<ChunkCoord, BufferedChunk>();
	
	private static Pool<ChunkBuffer> pool = new Pool<ChunkBuffer>(42);
	public static String getBufferSizeName() {
		int size = 80 * pool.size();
		if (size < 1000) {
			return size + " kb";
		} else {
			return ((double) size / 1000) + " mb";
		}
	}
	
	private static class ChunkCoord {
		public ChunkCoord(final World world, final int x, final int z) {
			this.world = world;
			this.x = x;
			this.z = z;
		}
		public final World world;
		public final int x;
		public final int z;
		public int hashCode() {
			int hash = 3;
	        hash = 21 * hash + this.world.hashCode();
	        hash = 21 * hash + (this.x ^ (this.x >> 16));
	        hash = 21 * hash + (this.z ^ (this.z >> 16));
	        return hash;
		}
		public boolean equals(Object o) {
			if (o == this) return true;
			if (o instanceof ChunkCoord) {
				ChunkCoord coord = (ChunkCoord) o;
				return coord.x == this.x && coord.z == this.z && coord.world == this.world;
			}
			return false;
		}
	}
	
	public static BufferedChunk removeUnloaded(final World world, final int x, final int z) {
		synchronized (unloadedChunks) {
			return unloadedChunks.remove(new ChunkCoord(world, x, z));
		}
	}
	
	public void markUnloadedSoon() {
		synchronized (unloadedChunks) {
			unloadedChunks.put(new ChunkCoord(this.world, this.x, this.z), this);
		}
	}
		
	public BufferedChunk(final World world, final int x, final int z) {
		super(world, x, z);
		this.buffer = pool.create(this);
		this.buffer.getData().apply(this);
	}
	
	private PoolBuffer<ChunkBuffer> buffer;
	@Override
	public ChunkBuffer getNewData() {
		return new ChunkBuffer();
	}
	
	public void setFree() {
		if (this.buffer != null) {
			this.buffer.setFree(this);
			this.buffer = null;
		}
	}
	
	public ChunkProviderServer getChunkProviderServer() {
		return ((WorldServer) this.world).chunkProviderServer;
	}
		
	public boolean isUnloaded() {
		return this.getChunkProviderServer().chunks.get(this.x, this.z) != this;
	}
	
	public String toString() {
		return "[" + this.world.getWorld().getName() + ", " + this.x + ", " + this.z + "]";
	}
	
	public CraftWorld getWorld() {
		return this.world.getWorld();
	}
	
	/*
	 * Fix to prevent index out of bounds for 'invalid' tile entities
	 */
	public void a(int i, int j, int k, TileEntity tileentity) {
		if (i < 0 || i > 15 || j < 0 || j > 127 || k < 0 || k > 15) {
			NoLagg.log(Level.WARNING, "Tile entity got stored wrongly, chunk " + this.toString() + " could be corrupted!");
			return;
		}
		super.a(i, j, k, tileentity);
	}
	
	public boolean containsCoordinates(int x, int y, int z) {
		return x >= 0 && x < 16 && y >= 0 && y < 128 && z >= 0 && z < 16;
	}
	
	@SuppressWarnings("rawtypes")
	public void clearData() {
		for (List list : this.entitySlices) list.clear();
		this.tileEntities.clear();
	}
	
	/*
	 * Prevents out of bounds exceptions for chunks
	 */
	public int getTypeId(int x, int y, int z) {
		if (!containsCoordinates(x, y, z)) {
			NoLagg.log(Level.WARNING, "Block type id was requested outside chunk bounds: [" + this.world.getWorld().getName() + ", " + (x + this.x << 4) + ", " + y + ", " + (z | this.z << 4) + "]");
			return 0;
		}
		return super.getTypeId(x, y, z);
	}
	
	public NBTTagCompound getBufferedTag(DataInputStream stream) throws Throwable {
		byte type = stream.readByte();
		if (type != 10) throw new IOException("Root tag must be a named compound tag");
		String name = stream.readUTF();
		//BufferedTag reading with buffers
		final Map<String, byte[]> buffers = new HashMap<String, byte[]>(5);
		buffers.put("Blocks", this.b);
		buffers.put("Data", this.g.a);
		buffers.put("SkyLight", this.h.a);
		buffers.put("BlockLight", this.i.a);
		buffers.put("HeightMap", this.heightMap);
		return (NBTTagCompound) NBTBaseBuffered.readFrom(type, name, buffers, stream);
	}
	public void load(BufferedChunkRegionLoader loader, NBTTagCompound comp) throws Throwable {
		this.load(loader, comp, comp.getCompound("Level"));
	}
	
	public void load(BufferedChunkRegionLoader loader, NBTTagCompound comp, NBTTagCompound level) throws Throwable {
		if (!level.hasKey("xPos")) level.setInt("xPos", this.x);
		if (!level.hasKey("zPos")) level.setInt("zPos", this.z);
		//do the coordinates match?
		int realX = level.getInt("xPos");
		int realZ = level.getInt("zPos");
		if (realX != this.x || realZ != this.z) {
			loader.relocate(this, comp, realX, realZ);
			throw new WrongChunkCoordinatesException();
		}
		
		
		this.b = level.getByteArray("Blocks");
		this.g = Util.getNibble(this.g, level.getByteArray("Data"), this.world.heightBits);
		this.h = Util.getNibble(this.h, level.getByteArray("SkyLight"), this.world.heightBits);
		this.i = Util.getNibble(this.i, level.getByteArray("BlockLight"), this.world.heightBits);
		
		this.heightMap = level.getByteArray("HeightMap");
		this.done = level.getBoolean("TerrainPopulated");
		if (!this.g.a()) {
			this.g = new NibbleArray(this.b.length, this.world.heightBits);
		}

		if (this.heightMap == null || !this.h.a()) {
			this.heightMap = new byte[256];
			this.h = new NibbleArray(this.b.length, this.world.heightBits);
			this.initLighting();
		}

		if (!this.i.a()) {
			this.i = new NibbleArray(this.b.length, this.world.heightBits);
			this.a();
		}

		NBTTagList nbttaglist = level.getList("Entities");

		if (nbttaglist != null) {
			for (int k = 0; k < nbttaglist.size(); ++k) {
				NBTTagCompound nbttagcompound1 = (NBTTagCompound) nbttaglist.get(k);
				Entity entity = EntityTypes.a(nbttagcompound1, this.world);

				this.s = true;
				if (entity != null) {
					this.a(entity);
				}
			}
		}

		NBTTagList nbttaglist1 = level.getList("TileEntities");

		if (nbttaglist1 != null) {
			for (int l = 0; l < nbttaglist1.size(); ++l) {
				NBTTagCompound nbttagcompound2 = (NBTTagCompound) nbttaglist1.get(l);
				TileEntity tileentity = TileEntity.c(nbttagcompound2);

				if (tileentity != null) {
					this.a(tileentity);
				}
			}
		}

		if (level.hasKey("TileTicks")) {
			NBTTagList nbttaglist2 = level.getList("TileTicks");

			if (nbttaglist2 != null) {
				for (int i1 = 0; i1 < nbttaglist2.size(); ++i1) {
					NBTTagCompound nbttagcompound3 = (NBTTagCompound) nbttaglist2.get(i1);

					this.world.d(nbttagcompound3.getInt("x"), nbttagcompound3.getInt("y"), nbttagcompound3.getInt("z"), nbttagcompound3.getInt("i"), nbttagcompound3.getInt("t"));
				}
			}
		}
	}

}